==> 0.4.1&file=in_ffsox-2a-0.4.1-src.tar.gz: in_ffsox-2a-0.4.1/in_ffsox/track.c <==
/*
 * track.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

track_t *track_init(track_t *t, const in_char *fn, prefs_t *ps, int full)
{
  //int mode=FFI_CSV|FFI_STREAM_META|FFI_AUDIO_STEREO;
  int mode=FFI_CSV|FFI_STREAM_META;
#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // [
  double drc_scale;
#endif // ]

  DMARKLN();
  DWPRINTF(L"track_init \"%s\"\n",fn);

  if (full)
    mode|=FFI_OPEN;

#if defined (FFSOX_AUDIO_STEREO) // [
  if (ps->audio.force_stereo)
    mode|=FFI_AUDIO_STEREO;
#endif // ]

  t->ps=ps;

  if (NULL==(t->fn=pb_str_cre(fn)))
    goto error;

#if defined (PREFS_AUDIO_DRC) // [
  if (ps->audio.drc.enabled) {
    mode|=FFI_AUDIO_DRC;
#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // [
    drc_scale=AUDIO_DRC_Q*ps->audio.drc.scale;
DWRITELN("-----------------------------------------------------");
DVWRITELN("drc.scale=%0.1f",drc_scale);
DWRITELN("-----------------------------------------------------");
#endif // ]
  }
#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // [
  else
    drc_scale=0.0;
#endif // ]
#endif // ]

#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // [
  if (NULL==ffi_init(&t->ffi,t->fn,mode,&drc_scale))
    goto error;
#else // ] [
  if (NULL==ffi_init(&t->ffi,t->fn,mode))
    goto error;
#endif // ]

  return t;
error:
  track_cleanup(t);

  return NULL;
}

void track_cleanup(track_t *t)
{
#define INFO_BEGIN(id) \
  if (NULL!=t->info.id) { \
    free(t->info.id); \
    t->info.id=NULL; \
  }
#define INFO_ALT(id,id2)
#define INFO_END()
#include "info.h"
#undef INFO_END
#undef INFO_ALT
#undef INFO_BEGIN

  ffi_cleanup(&t->ffi);
  t->ffi.ic=NULL;

  if (NULL!=t->fn) {
    free(t->fn);
    t->fn=NULL;
  }
}

char *track_get_tag(track_t *t, char *key)
{
  AVDictionaryEntry *entry=av_dict_get(ffi_metadata(&t->ffi),key,NULL,
        AV_DICT_DONT_STRDUP_KEY|AV_DICT_DONT_STRDUP_VAL);

  return NULL==entry||NULL==entry->value||0==*entry->value?NULL:entry->value;
}

double track_get_q(track_t *t)
{
  AVDictionaryEntry *track_gain=NULL;
  AVDictionaryEntry *album_gain=NULL;
  prefs_t *ps=t->ps;
  double album_db, track_db;
  double weight=ps->rg.weight;
  AVDictionary *metadata;

  if (ps->rg.off)
    return 0.0==ps->audio.preamp?1.0:DB2Q(ps->audio.preamp);
  else {
    metadata=ffi_metadata(&t->ffi);
    track_gain=av_dict_get(metadata,"replaygain_track_gain",NULL,
          AV_DICT_DONT_STRDUP_KEY|AV_DICT_DONT_STRDUP_VAL);
    album_gain=av_dict_get(metadata,"replaygain_album_gain",NULL,
          AV_DICT_DONT_STRDUP_KEY|AV_DICT_DONT_STRDUP_VAL);

    if (NULL==track_gain&&NULL==album_gain)
      return 0.0==ps->audio.preamp?1.0:DB2Q(ps->audio.preamp);
    else if (NULL==track_gain) {
      sscanf(album_gain->value,"%lf",&album_db);

      return DB2Q(ps->audio.preamp+album_db);
    }
    else if (NULL==album_gain) {
      sscanf(track_gain->value,"%lf",&track_db);

      return DB2Q(ps->audio.preamp+track_db);
    }
    else {
      sscanf(track_gain->value,"%lf",&track_db);
      sscanf(album_gain->value,"%lf",&album_db);

      return DB2Q(ps->audio.preamp+(weight*track_db+(1.0-weight)*album_db));
    }
  }
}

track_t *track_reopen(track_t *t, const in_char *fn, prefs_t *ps, int full)
{
#if defined (DEBUG) // {
  //fwprintf(stderr,L"track_reopen \"%s\"\n",fn);
#endif // }

  if (NULL!=t->ffi.ic&&0!=_tcscmp(PS_TCS(t->fn),fn)) {
    track_cleanup(t);
    memset(t,0,sizeof *t);
  }

  return NULL==t->ffi.ic&&NULL==track_init(t,fn,ps,full)?NULL:t;
}

void track_getfileinfo(track_t *t, in_char *title, int *length_in_ms)
{
  in_char *tp;
  in_char *fn,*fn1,*fn2;
  size_t len;

  if (length_in_ms)
    *length_in_ms=track_get_length(t);

  if (title) {
    tp=title;
    // get artist
    track_get_artist(t,tp,0);
    
    // if artist, append -
    if (title<(tp+=wcslen(tp))) {
      *tp++=L' ';
      *tp++=L'-';
      *tp++=L' ';
      *tp=0;
    }

    // get title
    track_get_title(t,tp,0);

    // if artist and no title, append file to artist
    if (title<tp&&0==*tp) {
      fn=PS_TCS(t->fn);
      fn2=fn+wcslen(fn);

      while (fn<fn2&&'.'!=*fn2)
        --fn2;

      fn1=fn2;

      while (fn<fn1&&'/'!=fn1[-1]&&'\\'!=fn1[-1])
        --fn1;

      len=fn2-fn1;
      memcpy(tp,fn1,len*sizeof *fn1);
      tp[len]=0;
    }
  }

#if defined (DEBUG) // {
  //fwprintf(stderr,L"track_getfileinfo: \"%s\", %d\n",title,*length_in_ms);
#endif // }
}

in_char *track_getfileinfo_ext(track_t *t, char *metadata, wchar_t *ret,
    int retlen)
{
#if defined (DEBUG) // {
  //fprintf(stderr,"track_getfileinfo_ext \"%s\"\n",metadata);
#endif // }
#if ! defined (RG_WEIGHT_BOX) // {
  prefs_t *ps=t->ps;
#endif // }

  if (0==stricmp("length",metadata)) {
    _snwprintf(ret,retlen,L"%d",track_get_length(t));
    ret[retlen-1]=0;
    return ret;
  }
  else if (0==stricmp("type",metadata)) {
    _snwprintf(ret,retlen,L"%d",NULL==t->ffi.vst?0:1);
    ret[retlen-1]=0;
    DPRINTF("%s: ",metadata);
    DWPRINTF(L"%s\n",ret);
    return ret;
  }
  else if (0==stricmp("formatinformation",metadata))
    return track_get_formatinformation(t,ret,retlen);
#if ! defined (RG_WEIGHT_BOX) // {
  else if (!ps->rg.off&&ps->rg.comment&&0==stricmp("comment",metadata))
    return track_get_overwrite_comment(t,ret,retlen);
#else // } {
  else if (0!=t->ps->rg.comment&&0==stricmp("comment",metadata))
    return track_get_overwrite_comment(t,ret,retlen);
#endif // }
#define INFO_BEGIN(id) \
  else if (0==stricmp(#id,metadata)) \
    return track_get_##id(t,ret,retlen);
#define INFO_ALT(id,id2)
#define INFO_END()
#include "info.h"
#undef INFO_END
#undef INFO_ALT
#undef INFO_BEGIN

#if defined (DEBUG_TAGS) // {
  DPRINTF("\"%s\"\n",metadata);
#endif // }

  return NULL;
}

int track_get_length(track_t *t)
{
  return NULL==t->ffi.ic?-1000:round((double)t->ffi.ic->duration/1000.0);
}

in_char *track_get_overwrite_comment(track_t *t, in_char *buf, size_t n)
{
  wbuf_t wb;
  double weight=t->ps->rg.weight;
  double gain_q=track_get_q(t);
  char *peak=track_get_tag(t,"replaygain_track_peak");
  double peak_q;

  wbuf_init(&wb,buf,n);

  if (0.0<=weight) {
    wbuf_add_wva(&wb,L"Preamp: %.02lf dB",t->ps->audio.preamp);
    wbuf_add_wva(&wb,L", Weight: %.0lf%%",100.0*t->ps->rg.weight);
    wbuf_add_wva(&wb,L"\r\nApplied Gain: %+.02lf dB",Q2DB(gain_q));

    if (NULL!=peak&&1==sscanf(peak,"%lf",&peak_q)) {
      wbuf_add_wva(&wb,L"\r\nPeak: %+.02lf dBFS -> %+.02lf dBFS",Q2DB(peak_q),
          Q2DB(gain_q*peak_q));
    }
  }
  else if (NULL!=peak&&1==sscanf(peak,"%lf",&peak_q))
    wbuf_add_wva(&wb,L"Peak: %+.02lf",Q2DB(peak_q));

  return buf;
}

static int track_get_input_rate(track_t *t)
{
  ffi_t *ffi=&t->ffi;
  ffi_st_t *ast=ffi->ast;
  AVCodecContext *dc=NULL==ast?NULL:ast->dc;
  int rate;

  if (NULL==dc)
    return 0;
  else {
    dc=ffi->ast->dc;
    rate=dc->sample_rate;

    return rate;
  }
}

static int track_get_output_rate(track_t *t)
{
  prefs_t *ps=t->ps;
  int input_rate=track_get_input_rate(t);
  int output_rate=ps->sox.rate;

  if (ps_sox_mode_off(ps))
    return input_rate;
  else if (0==output_rate)
    return input_rate;
  else if (ps->sox.rate_min&&output_rate<=input_rate)
    return input_rate;
  else if (ps->sox.rate_mul) {
    while (input_rate<output_rate)
      input_rate<<=1;

    return input_rate;
  }
  else
    return output_rate;
}

static int track_get_input_bps(track_t *t)
{
  ffi_t *ffi=&t->ffi;
  ffi_st_t *ast=ffi->ast;
  AVCodecContext *dc=ast->dc;

  return NULL==(dc)?0:
      0<(dc)->bits_per_raw_sample?(dc)->bits_per_raw_sample:
      8*av_get_bytes_per_sample((dc)->sample_fmt);
}

static int track_get_output_bps(track_t *t)
{
  return t->ps->general.bps;
}

int track_sox_needed(track_t *t, ap_t *in, ap_t *out)
{
  prefs_t *ps=t->ps;
  ffi_t *ffi=&t->ffi;
  ffi_st_t *ast=ffi->ast;
  AVCodecContext *dc=ast->dc;
  int sox_needed;

  in->rate=track_get_input_rate(t);
  in->nch=NULL==dc?2:dc->channels;
  in->bps=track_get_input_bps(t);

  out->rate=track_get_output_rate(t);
  out->nch=NULL==dc?2:dc->channels;
  out->bps=track_get_output_bps(t);

  sox_needed=ps_sox_mode_off(ps)?0:
      ps_sox_mode_on(ps)?1:
      out->rate==in->rate?0:
      1;

  if (!sox_needed)
    out->rate=in->rate;

  return sox_needed;
}

in_char *track_get_formatinformation(track_t *t, in_char *buf, size_t n)
{
  ffi_t *ffi=&t->ffi;
  ffi_st_t *ast=ffi->ast;
  ffi_st_t *vst=ffi->vst;
  wbuf_t wb;
  int length,h,min,s;
  ap_t in,out;

  wbuf_init(&wb,buf,n);
  wbuf_add(&wb,L"Decoder: FFmpeg");

  length=track_get_length(t)/1000;
  s=length%60;
  length/=60;
  min=length%60;
  length/=60;
  h=length;

  if (0<h)
    wbuf_add_wva(&wb,L"\nLength: %d:%02d:%02d",h,min,s);
  else
    wbuf_add_wva(&wb,L"\nLength: %d:%02d",min,s);

  wbuf_add_va(&wb,"\n\nAudio codec: %s",ast->dec->name);
  track_sox_needed(t,&in,&out);

  if (0<ast->dc->bit_rate)
    wbuf_add_wva(&wb,L"\nBitrate: %d kb/s",ast->dc->bit_rate/1000);

  wbuf_add_wva(&wb,L"\nBits per Sample (in): %d",in.bps);
  wbuf_add_wva(&wb,L"\nSample Rate (in): %d Hz",in.rate);
  wbuf_add_wva(&wb,L"\nBits per Sample (out): %d",out.bps);
  wbuf_add_wva(&wb,L"\nSample Rate (out): %d Hz",out.rate);

  if (0!=t->ps->general.vid&&NULL!=vst)
    wbuf_add_va(&wb,"\n\nVideo codec: %s",vst->dec->name);

  return buf;
}

static in_char *track_get_value(track_t *t, pb_str_t **val, char *key,
    in_char *buf, size_t n)
{
  in_char *state=NULL;
  AVDictionaryEntry *tag=NULL;

  if (NULL==*val) {
    if (NULL==(tag=av_dict_get(ffi_metadata(&t->ffi),key,NULL,
        AV_DICT_DONT_STRDUP_KEY|AV_DICT_DONT_STRDUP_VAL)))
      goto cleanup;

    if (NULL==(*val=pb_str_cre_cs(tag->value)))
      goto cleanup;
  }

  if (0<n) {
    _tcsncpy(buf,(*val)->tcs,n);
    buf[n-1]=_T('\0');
  }
  else
    _tcscpy(buf,(*val)->tcs);

  state=buf;
cleanup:
  return state;
}

#define INFO_BEGIN(id) \
in_char *track_get_##id(track_t *t, in_char *buf, size_t n) \
{ \
  if (NULL!=track_get_value(t,&t->info.id,#id,buf,n)) \
    return buf;
#define INFO_ALT(id,id2) \
  else if (NULL!=track_get_value(t,&t->info.id,#id2,buf,n)) \
    return buf;
#define INFO_END() \
  return buf; \
}
#include "info.h"
#undef INFO_END
#undef INFO_ALT
#undef INFO_BEGIN

==> 0.4.2&file=in_ffsox-2a-0.4.2-src.tar.gz: in_ffsox-2a-0.4.2/in_ffsox/track.c <==
/*
 * track.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

track_t *track_init(track_t *t, const in_char *fn, prefs_t *ps, int full)
{
  //int mode=FFI_CSV|FFI_STREAM_META|FFI_AUDIO_STEREO;
  int mode=FFI_CSV|FFI_STREAM_META;
#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // [
  double drc_scale;
#endif // ]

  DMARKLN();
  DWPRINTF(L"track_init \"%s\"\n",fn);

  if (full)
    mode|=FFI_OPEN;

#if defined (FFSOX_AUDIO_STEREO) // [
  if (ps->audio.force_stereo)
    mode|=FFI_AUDIO_STEREO;
#endif // ]

  t->ps=ps;

  if (NULL==(t->fn=pb_str_cre(fn)))
    goto error;

#if defined (PREFS_AUDIO_DRC) // [
  if (ps->audio.drc.enabled) {
    mode|=FFI_AUDIO_DRC;
#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // [
    drc_scale=AUDIO_DRC_Q*ps->audio.drc.scale;
#endif // ]
  }
#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // [
  else
    drc_scale=0.0;
#endif // ]
#endif // ]

#if /*! defined (FFSOX_AUDIO_CUST) &&*/ defined (FFSOX_AUDIO_DRC) // [
  if (NULL==ffi_init(&t->ffi,t->fn,mode,&drc_scale)) {
    DMESSAGE("ffi_init() failed");
    goto error;
  }
#else // ] [
  if (NULL==ffi_init(&t->ffi,t->fn,mode)) {
    DMESSAGE("ffi_init() failed");
    goto error;
  }
#endif // ]

  return t;
error:
  track_cleanup(t);

  return NULL;
}

void track_cleanup(track_t *t)
{
#define INFO_BEGIN(id) \
  if (NULL!=t->info.id) { \
    free(t->info.id); \
    t->info.id=NULL; \
  }
#define INFO_ALT(id,id2)
#define INFO_END()
#include "info.h"
#undef INFO_END
#undef INFO_ALT
#undef INFO_BEGIN

  ffi_cleanup(&t->ffi);
  t->ffi.ic=NULL;

  if (NULL!=t->fn) {
    free(t->fn);
    t->fn=NULL;
  }
}

char *track_get_tag(track_t *t, char *key)
{
  AVDictionaryEntry *entry=av_dict_get(ffi_metadata(&t->ffi),key,NULL,
        AV_DICT_DONT_STRDUP_KEY|AV_DICT_DONT_STRDUP_VAL);

  return NULL==entry||NULL==entry->value||0==*entry->value?NULL:entry->value;
}

double track_get_q(track_t *t)
{
  AVDictionaryEntry *track_gain=NULL;
  AVDictionaryEntry *album_gain=NULL;
  prefs_t *ps=t->ps;
  double album_db, track_db;
  double weight=ps->rg.weight;
  AVDictionary *metadata;

  if (ps->rg.off)
    return 0.0==ps->audio.preamp?1.0:DB2Q(ps->audio.preamp);
  else {
    metadata=ffi_metadata(&t->ffi);
    track_gain=av_dict_get(metadata,"replaygain_track_gain",NULL,
          AV_DICT_DONT_STRDUP_KEY|AV_DICT_DONT_STRDUP_VAL);
    album_gain=av_dict_get(metadata,"replaygain_album_gain",NULL,
          AV_DICT_DONT_STRDUP_KEY|AV_DICT_DONT_STRDUP_VAL);

    if (NULL==track_gain&&NULL==album_gain)
      return 0.0==ps->audio.preamp?1.0:DB2Q(ps->audio.preamp);
    else if (NULL==track_gain) {
      sscanf(album_gain->value,"%lf",&album_db);

      return DB2Q(ps->audio.preamp+album_db);
    }
    else if (NULL==album_gain) {
      sscanf(track_gain->value,"%lf",&track_db);

      return DB2Q(ps->audio.preamp+track_db);
    }
    else {
      sscanf(track_gain->value,"%lf",&track_db);
      sscanf(album_gain->value,"%lf",&album_db);

      return DB2Q(ps->audio.preamp+(weight*track_db+(1.0-weight)*album_db));
    }
  }
}

track_t *track_reopen(track_t *t, const in_char *fn, prefs_t *ps, int full)
{
#if defined (DEBUG) // {
  //fwprintf(stderr,L"track_reopen \"%s\"\n",fn);
#endif // }

  if (NULL!=t->ffi.ic&&0!=_tcscmp(PS_TCS(t->fn),fn)) {
    track_cleanup(t);
    memset(t,0,sizeof *t);
  }

  return NULL==t->ffi.ic&&NULL==track_init(t,fn,ps,full)?NULL:t;
}

void track_getfileinfo(track_t *t, in_char *title, int *length_in_ms)
{
  in_char *tp;
  in_char *fn,*fn1,*fn2;
  size_t len;

  if (length_in_ms)
    *length_in_ms=track_get_length(t);

  if (title) {
    tp=title;
    // get artist
    track_get_artist(t,tp,0);
    
    // if artist, append -
    if (title<(tp+=wcslen(tp))) {
      *tp++=L' ';
      *tp++=L'-';
      *tp++=L' ';
      *tp=0;
    }

    // get title
    track_get_title(t,tp,0);

    // if artist and no title, append file to artist
    if (title<tp&&0==*tp) {
      fn=PS_TCS(t->fn);
      fn2=fn+wcslen(fn);

      while (fn<fn2&&'.'!=*fn2)
        --fn2;

      fn1=fn2;

      while (fn<fn1&&'/'!=fn1[-1]&&'\\'!=fn1[-1])
        --fn1;

      len=fn2-fn1;
      memcpy(tp,fn1,len*sizeof *fn1);
      tp[len]=0;
    }
  }

#if defined (DEBUG) // {
  //fwprintf(stderr,L"track_getfileinfo: \"%s\", %d\n",title,*length_in_ms);
#endif // }
}

in_char *track_getfileinfo_ext(track_t *t, char *metadata, wchar_t *ret,
    int retlen)
{
#if defined (DEBUG) // {
  //fprintf(stderr,"track_getfileinfo_ext \"%s\"\n",metadata);
#endif // }
#if ! defined (RG_WEIGHT_BOX) // {
  prefs_t *ps=t->ps;
#endif // }

  if (0==stricmp("length",metadata)) {
    _snwprintf(ret,retlen,L"%d",track_get_length(t));
    ret[retlen-1]=0;
    return ret;
  }
  else if (0==stricmp("type",metadata)) {
    _snwprintf(ret,retlen,L"%d",NULL==t->ffi.vst?0:1);
    ret[retlen-1]=0;
    DPRINTF("%s: ",metadata);
    DWPRINTF(L"%s\n",ret);
    return ret;
  }
  else if (0==stricmp("formatinformation",metadata))
    return track_get_formatinformation(t,ret,retlen);
#if ! defined (RG_WEIGHT_BOX) // {
  else if (!ps->rg.off&&ps->rg.comment&&0==stricmp("comment",metadata))
    return track_get_overwrite_comment(t,ret,retlen);
#else // } {
  else if (0!=t->ps->rg.comment&&0==stricmp("comment",metadata))
    return track_get_overwrite_comment(t,ret,retlen);
#endif // }
#define INFO_BEGIN(id) \
  else if (0==stricmp(#id,metadata)) \
    return track_get_##id(t,ret,retlen);
#define INFO_ALT(id,id2)
#define INFO_END()
#include "info.h"
#undef INFO_END
#undef INFO_ALT
#undef INFO_BEGIN

#if defined (DEBUG_TAGS) // {
  DPRINTF("\"%s\"\n",metadata);
#endif // }

  return NULL;
}

int track_get_length(track_t *t)
{
  return NULL==t->ffi.ic?-1000:round((double)t->ffi.ic->duration/1000.0);
}

in_char *track_get_overwrite_comment(track_t *t, in_char *buf, size_t n)
{
  wbuf_t wb;
  double weight=t->ps->rg.weight;
  double gain_q=track_get_q(t);
  char *peak=track_get_tag(t,"replaygain_track_peak");
  double peak_q;

  wbuf_init(&wb,buf,n);

  if (0.0<=weight) {
    wbuf_add_wva(&wb,L"Preamp: %.02lf dB",t->ps->audio.preamp);
    wbuf_add_wva(&wb,L", Weight: %.0lf%%",100.0*t->ps->rg.weight);
    wbuf_add_wva(&wb,L"\r\nApplied Gain: %+.02lf dB",Q2DB(gain_q));

    if (NULL!=peak&&1==sscanf(peak,"%lf",&peak_q)) {
      wbuf_add_wva(&wb,L"\r\nPeak: %+.02lf dBFS -> %+.02lf dBFS",Q2DB(peak_q),
          Q2DB(gain_q*peak_q));
    }
  }
  else if (NULL!=peak&&1==sscanf(peak,"%lf",&peak_q))
    wbuf_add_wva(&wb,L"Peak: %+.02lf",Q2DB(peak_q));

  return buf;
}

static int track_get_input_rate(track_t *t)
{
  ffi_t *ffi=&t->ffi;
  ffi_st_t *ast=ffi->ast;
  AVCodecContext *dc=NULL==ast?NULL:ast->dc;
  int rate;

  if (NULL==dc)
    return 0;
  else {
    dc=ffi->ast->dc;
    rate=dc->sample_rate;

    return rate;
  }
}

static int track_get_output_rate(track_t *t)
{
  prefs_t *ps=t->ps;
  int input_rate=track_get_input_rate(t);
  int output_rate=ps->sox.rate;

  if (ps_sox_mode_off(ps))
    return input_rate;
  else if (0==output_rate)
    return input_rate;
  else if (ps->sox.rate_min&&output_rate<=input_rate)
    return input_rate;
  else if (ps->sox.rate_mul) {
    while (input_rate<output_rate)
      input_rate<<=1;

    return input_rate;
  }
  else
    return output_rate;
}

static int track_get_input_bps(track_t *t)
{
  ffi_t *ffi=&t->ffi;
  ffi_st_t *ast=ffi->ast;
  AVCodecContext *dc=ast->dc;

  return NULL==(dc)?0:
      0<(dc)->bits_per_raw_sample?(dc)->bits_per_raw_sample:
      8*av_get_bytes_per_sample((dc)->sample_fmt);
}

static int track_get_output_bps(track_t *t)
{
  return t->ps->general.bps;
}

int track_sox_needed(track_t *t, ap_t *in, ap_t *out)
{
  prefs_t *ps=t->ps;
  ffi_t *ffi=&t->ffi;
  ffi_st_t *ast=ffi->ast;
  AVCodecContext *dc=ast->dc;
  int sox_needed;

  in->rate=track_get_input_rate(t);
  in->nch=NULL==dc?2:dc->channels;
  in->bps=track_get_input_bps(t);

  out->rate=track_get_output_rate(t);
  out->nch=NULL==dc?2:dc->channels;
  out->bps=track_get_output_bps(t);

  sox_needed=ps_sox_mode_off(ps)?0:
      ps_sox_mode_on(ps)?1:
      out->rate==in->rate?0:
      1;

  if (!sox_needed)
    out->rate=in->rate;

  return sox_needed;
}

in_char *track_get_formatinformation(track_t *t, in_char *buf, size_t n)
{
  ffi_t *ffi=&t->ffi;
  ffi_st_t *ast=ffi->ast;
  ffi_st_t *vst=ffi->vst;
  wbuf_t wb;
  int length,h,min,s;
  ap_t in,out;

  wbuf_init(&wb,buf,n);
  wbuf_add(&wb,L"Decoder: FFmpeg");

  length=track_get_length(t)/1000;
  s=length%60;
  length/=60;
  min=length%60;
  length/=60;
  h=length;

  if (0<h)
    wbuf_add_wva(&wb,L"\nLength: %d:%02d:%02d",h,min,s);
  else
    wbuf_add_wva(&wb,L"\nLength: %d:%02d",min,s);

  wbuf_add_va(&wb,"\n\nAudio codec: %s",ast->dec->name);
  track_sox_needed(t,&in,&out);

  if (0<ast->dc->bit_rate)
    wbuf_add_wva(&wb,L"\nBitrate: %d kb/s",ast->dc->bit_rate/1000);

  wbuf_add_wva(&wb,L"\nBits per Sample (in): %d",in.bps);
  wbuf_add_wva(&wb,L"\nSample Rate (in): %d Hz",in.rate);
  wbuf_add_wva(&wb,L"\nBits per Sample (out): %d",out.bps);
  wbuf_add_wva(&wb,L"\nSample Rate (out): %d Hz",out.rate);

#if defined (VIS_DELAY) // [
  if (0!=t->ps->general.vid.enabled&&NULL!=vst)
    wbuf_add_va(&wb,"\n\nVideo codec: %s",vst->dec->name);
#else // ] [
  if (0!=t->ps->general.vid&&NULL!=vst)
    wbuf_add_va(&wb,"\n\nVideo codec: %s",vst->dec->name);
#endif // ]

  return buf;
}

static in_char *track_get_value(track_t *t, pb_str_t **val, char *key,
    in_char *buf, size_t n)
{
  in_char *state=NULL;
  AVDictionaryEntry *tag=NULL;

  if (NULL==*val) {
    if (NULL==(tag=av_dict_get(ffi_metadata(&t->ffi),key,NULL,
        AV_DICT_DONT_STRDUP_KEY|AV_DICT_DONT_STRDUP_VAL)))
      goto cleanup;

    if (NULL==(*val=pb_str_cre_cs(tag->value)))
      goto cleanup;
  }

  if (0<n) {
    _tcsncpy(buf,(*val)->tcs,n);
    buf[n-1]=_T('\0');
  }
  else
    _tcscpy(buf,(*val)->tcs);

  state=buf;
cleanup:
  return state;
}

#define INFO_BEGIN(id) \
in_char *track_get_##id(track_t *t, in_char *buf, size_t n) \
{ \
  if (NULL!=track_get_value(t,&t->info.id,#id,buf,n)) \
    return buf;
#define INFO_ALT(id,id2) \
  else if (NULL!=track_get_value(t,&t->info.id,#id2,buf,n)) \
    return buf;
#define INFO_END() \
  return buf; \
}
#include "info.h"
#undef INFO_END
#undef INFO_ALT
#undef INFO_BEGIN
